#include <Servo.h>

/* Multi Servo Switch Sofware V1.0 (Stand 02.08.2020)
 * Ersteller P. Doggwiler
 * Datum: 14.05.2020
 * Über einen Proportional Kanal oder Taster können bis zu 10 Ausgänge mit einem PWM Signal ab dem Sender gesteuert werden. 
 * Über einen Counter wird zum entsprechenden Ausgang hochgezählt, das an Pin13 anliegende Signal wird danach an den Ausgang weitergeleitet.
 * Einsatz der Printplatte: 
 * Die Ausgänge müssen über die Jumper direkt auf den Ausgang geschaltet werden.
 * Digital Input 2 ist für den Counter
 * Digital Input 13 ist für das PWM Signal ab Empfänger
 * Alle Ausgangssignal werden bei der Initialisierung auf Wert 1500 gesetzt (Mittelstellung)
 * 
 * Getestet mit Sender/Empfänger
 * Graupner HOTT
 * Carson Multireflex
 * 
 * 
 * <<<<<<<<<<<<<<<<<<<<<<!!!!!!!! Start Wichtige Hinweise (02.08.2020) !!!!!!!!>>>>>>>>>>>>>>>>>>>>>>
 * Test haben ergeben, das bei gewissen Servos das PWM Signal beim Steuern das Signal auf den Emfpägerausgang überspringt, 
 * welches zum MulitSwitch Print führt. Dieses Siganlüberspringen wird eingelesen und Zählt den Counter hoch.
 * Einzige Lösung bisher ist der Austausch des Servos.
 * Das Problem konnte mangels Alternativen an Fernsteuerungen/Empfängern nur mit GRAUPNER HOTT Anlagen getestet und nachvollzogen werden.
 * <<<<<<<<<<<<<<<<<<<<<<!!!!!!!! Ende wichtige Hinweise !!!!!!!!>>>>>>>>>>>>>>>>>>>>>>
 * 
 * 
 * Änderungen V1.1
    * Datum 02.08.2020
    * Code Umgeschrieben, unnötige Komentare gelöscht 
 *
 *
 *
 * Änderungen V1.2
    * Datum: 21.11.2020
    * Erweiterung des Codes damit dieser als MultiServo oder als MultiUmpolmodul Baustein genutzt werden kann
    * Eingang A5 auf VCC (Spannung Arduino) => MultiUmpolmdoul. Es können über 4 Analoge Eingänge fünf Umpolrelais (zwei pro Eingang) direkt geschaltet werden
    * Eingang A5 auf 0V => MultiServo aktiv
    * Alle nicht benötigetn analogen Eingänge müssen auf GND gestellt werden!!
 *
 *
 */




/*********************************************************
 * Start des Programms mit den Definierungen der Eingänge
 *********************************************************/
#define PinOut1 3
#define PinOut2 4
#define PinOut3 5
#define PinOut4 6
#define PinOut5 7
#define PinOut6 8
#define PinOut7 9
#define PinOut8 10
#define PinOut9 11
#define PinOut10 12

#define PinInA1 A1
#define PinInA2 A2
#define PinInA3 A3
#define PinInA4 A4
#define PinInA5 A5


#define PinOutStart A0       //Blinkende LED im letzten Schritt der Setup Routine  => Ändern auf PIN ID A0 (Analog Output 0), ist die LED auf dem JoyIt Board


//#define DEBUG          // Serial output setup
//#define DEBUG_PWM 



//Servo Objekt Erstellen
Servo myservo;  // Servo-Objekt aus Library


const int Ausgaenge_Array[10] = {PinOut1, PinOut2, PinOut3, PinOut4, PinOut5, PinOut6, PinOut7, PinOut8, PinOut9, PinOut10};


/***********************************
 * Erweiterungen für MultiServo
 ***********************************/
 
volatile const int PinIn2 = 2;                        //Eingang für Counter Signal oder für Auswahl ob MultiServo = LOW oder Relais Umpolmodul für Motoren = HIGH
volatile const int PinIn13 = 13;                      //Eingang für PWM Signal

unsigned long getPWM_Input = 0;                       //Eingelesendes PWM Signal
unsigned long getPWM_Output = 0;
volatile int counter = 0;                             //Counter für Ausgang Kanal schalten


//Zeit Variablen für Ausgang schalten
unsigned long aktZeit = 0;                            //Aktuelle Zeit
volatile unsigned long alteZeit = 0;                  //Zeit für die Deltaberechnung und Freigabe des Ausgangs
long delta = 0;                                       //Delta Zeit = aktZeit -  =>  muss grösser sein als die MultiSwitchZeit und der Ausgang wird geschaltet
volatile unsigned long MultiSwitchZeit = 1500;        //Zeit welche es keinen Wechsel mehr gibt und der index grösser 8 ist


//Zeit Variablen für Entprellen des PWM Signals
unsigned long aktZeit_PWM = 0;                        //Aktuelle Zeit
volatile unsigned long alteZeit_PWM = 0;              //Delta Zeit für Freigabe
volatile unsigned long entPrell_PWM = 10;             //Zeit für Prellen der Taster
volatile bool ErsterHighPuls = false;                 //Erster PWM Wert > 1800 oder <1200 (geht beides)
volatile bool FreigabeAusgang = false;                //Freigabe für den Ausgang, erfolgt erst wenn Taster losgelassen wird
volatile bool FreigabeAusgangReset = false;           //Freigabe für die Rücksetzung des Counters
volatile bool SetInOutPWM = false;                    //Freigabe damit PWM im Loop vom Eingang an Ausgang übermittelt wird

//Variable für PWM Signal
volatile unsigned long PWM_Start_Input = 1500;        //Servo ist auf Mitelstellung




/***********************************
 * Erweiterungen für MultiUmpolmodul
 ***********************************/
volatile bool SetModulTyp = false;                    //False ist MultiServo, True ist MultiUmpolmodul
const int Eingang_An_Array[4] = {PinInA1, PinInA2, PinInA3, PinInA4};
volatile int Eingang_An_Status[4] = {4, 4, 4, 4};  //0 => Mittelposition | 1 => neg. Ausschlag |  2 => pos. Ausschlag | 4 => Kein Signal am Eingang







void setup() 
{

  /**************************************************************
   * Wartezeit von 10s bis das Setup Programm ausgeführt wird.
   **************************************************************/
  long waitTime = millis();
  int WaitingTime = 10000;        //Wartezeit in Millisekunden => Default sind 10sek
  int waitTimeTemp;
  
  while(waitTimeTemp <= WaitingTime)
  {
    waitTimeTemp = millis() - waitTime;  
  }


  
  //Notwendig wenn die Ausgabe auf dem Seriellen Monitor für Debug Zwecke erfolgen soll
  // Serial.begin(115200);

   
   #ifdef DEBUG        // Serial output on 
   Serial.println("<<<<<<<<<<<<<<<<<<<< Start Setup Routine! >>>>>>>>>>>>>>>>>>>>>>");
   #endif

  pinMode(PinInA5, INPUT_PULLUP);              //Eingang initialisierne
  //Interner Pull up Widerstand auf 5V, auf 0V setzen
  digitalWrite(PinInA5, LOW);                  //Innenwiederstand zurücksetzen


  //PinIn2: Auslesen für Wahl des Programms, PinIn2 wird danach ebenfalls auch für den Counter beim MultiServo eingesetzt
  if(digitalRead(PinInA5) == HIGH)
  {
      /******************************
       * Setup für UmpolModul
       ******************************/
       #ifdef DEBUG        // Serial output on 
      Serial.println("Setup UmpolModul Routine!");
      #endif

      SetModulTyp = true;
      
      //Analoge Pins als Ausgänge definieren
      for(int i = 0; i < 4; i++)
      {
        pinMode(Eingang_An_Array[i], INPUT_PULLUP);
        digitalWrite(Eingang_An_Array[i], LOW);  
      }

      //Definieren welche Eingänge aktiv (PWM Signal am Pin) sind, nur diese werden später eingelesen
      //Wert 4 ist Status inaktiv und am Eingang kein Signal anliegend
      for(int xx = 0; xx < 4; xx++)
      {
        if(pulseIn(Eingang_An_Array[xx], HIGH) > 10)      //PWM Signal liegt an
          Eingang_An_Status[xx] = 0;

        #ifdef DEBUG        // Serial output on 
          Serial.println(pulseIn(Eingang_An_Array[xx], HIGH));
        //else
          //Eingang_An_Status[xx] = 4;   
         #endif         
      }
      
  
      //Ausgänge Initialisieren
      for(int ii = 0; ii < 8; ii++)
      {
          pinMode(Ausgaenge_Array[ii], OUTPUT);
          digitalWrite(Ausgaenge_Array[ii], LOW);
      }

      #ifdef DEBUG        // Serial output on 
        Serial.println("Start Eingang_An_Status Init");
        for(int i=0; i<4;i++)
        {

            if(Eingang_An_Status[i] != 4)
            Serial.print("A");
            else
            Serial.print("I");
            Serial.print(" | ");               
        }
        Serial.println("Ende Eingang_An_Status Init");    
        #endif  
    
  }



  
  if(digitalRead(PinInA5) == LOW)
  {

      /******************************
       *MultiServo
      ******************************/
      #ifdef DEBUG        // Serial output on 
        Serial.println("Setup MultiServo Routine!"); 
      #endif
        
        
        //Eingänge Initialisieren  
        pinMode(PinIn2, INPUT_PULLUP);
        //pinMode(PinIn13, INPUT_PULLUP);
      
        pinMode(PinOutStart, OUTPUT);               //Ausgang für Sichtbarkeit das Programm gestartet ist. Analoger Ausgang als Digital Output setzen
        digitalWrite(PinIn2, LOW);              //Counter bzw. Taster, Interner Pull up Widerstand auf 5V, auf 0V setzen
        //digitalWrite(PinIn13, LOW);               //PWM Signal

  
      //Servo Ausgänge auf Mittelstellung setzen
      for(int i = 3; i < 13; i++)
      {
          myservo.attach(i);     // Ausgang als Servo Signal setzen
          myservo.writeMicroseconds(PWM_Start_Input);   // Servo auf Mittelpunkt setzen
          
      } 
          myservo.detach();     // Ausgang zürücksetzen (Wird jeweils beim Aufruf des Augangs als Servo Ausgang neu gesetzt.


  }



  /*****************************************************************************
  Start Kontrollausgabe nach dem Setup
  *****************************************************************************/
      
  digitalWrite(PinOutStart, HIGH);
  delay(1000);
  digitalWrite(PinOutStart, LOW);
  delay(1000);
  for(int i = 0; i < 5; i++)
  {
    digitalWrite(PinOutStart, HIGH);
    delay(200); 
    digitalWrite(PinOutStart, LOW);
    delay(200);
  }

  digitalWrite(PinOutStart, HIGH);
  delay(2000);
  digitalWrite(PinOutStart, LOW);

  #ifdef DEBUG        // Serial output on 
  Serial.println("<<<<<<<<<<<<<<<<<<<< Ende Setup Routine! >>>>>>>>>>>>>>>>>>>>>>");
  #endif
  /*****************************************************************************
  Ende Kontrollausgabe nach dem Setup
  *****************************************************************************/

  
}



void loop() {



  if(SetModulTyp == true)
  {
      /*************************************
       * Beginn MultiUmpolmodul
       *************************************/
      
      
      while(true)
      {
        //Status Array: 0 => Mittelposition - 1 => neg. Ausschlag -  2 => pos. Ausschlag
        int tAusg_1 = 0;
        int tAusg_2 = 0;  

        for(int iii = 0; iii < 4; iii++)
        {

              /*****************************************************
               * Nur wenn eingang gesetzt wird eingelsen
               * Switch Case ist in jeder klammer drin, damit die Ausgangsbestimmung 
               * nur dann geschieht wenn eine Änderung am Eingang vorhanden ist.
               * Beide Relais werden vor der Umpolung zuerst auf inaktiv geschaltet, um einen Kurzschluss zu vermeiden
               *****************************************************/
               
              if(Eingang_An_Status[iii] != 4)                             //Abfrage ob am entsprechenden Eingang ein Signal vorhanden war beim Initialisieren
              {

                getPWM_Input = pulseIn(Eingang_An_Array[iii], HIGH);      //PWM Signal am Eingang entsprechendne Eingang einlesen


                
                #ifdef DEBUG_PWM        // Serial output on 
                Serial.print("PWM Input Signal: E");
                Serial.print(iii);
                Serial.print(" | Wert");
                Serial.println(getPWM_Input);
                
                #endif
                

                
                if(getPWM_Input < 1200 && Eingang_An_Status[iii] != 1)
                {
                    switch(iii)
                    {
                      case 0:
                        tAusg_1 = 0;
                        tAusg_2 = 1;
                       break;
                      case 1:
                        tAusg_1 = 2;
                        tAusg_2 = 3;
                       break;
                      case 2:
                        tAusg_1 = 4;
                        tAusg_2 = 5;
                       break;
                      case 3:
                        tAusg_1 = 6;
                        tAusg_2 = 7;
                       break;
                      case 4:
                        tAusg_1 = 8;
                        tAusg_2 = 9;
                       break;
                      default:
                        tAusg_1 = 0;
                        tAusg_2 = 1;
                       break;                                                        
                    }
       
                    //Zuerst wird aus Sicherheitsgründen beide Relais ausgeschaltet
                    digitalWrite(Ausgaenge_Array[tAusg_1], LOW);
                    digitalWrite(Ausgaenge_Array[tAusg_2], LOW);                
                    digitalWrite(Ausgaenge_Array[tAusg_2], HIGH);

                    Eingang_An_Status[iii] = 1;

                    #ifdef DEBUG        // Serial output on 
                    Serial.println(" | Status: 1");
                    #endif
                    

                }
    
                if(getPWM_Input > 1800 && Eingang_An_Status[iii] != 2)
                {
                    switch(iii)
                    {
                      case 0:
                        tAusg_1 = 0;
                        tAusg_2 = 1;
                       break;
                      case 1:
                        tAusg_1 = 2;
                        tAusg_2 = 3;
                       break;
                      case 2:
                        tAusg_1 = 4;
                        tAusg_2 = 5;
                       break;
                      case 3:
                        tAusg_1 = 6;
                        tAusg_2 = 7;
                       break;
                      case 4:
                        tAusg_1 = 8;
                        tAusg_2 = 9;
                       break;
                      default:
                        tAusg_1 = 0;
                        tAusg_2 = 1;
                       break;                                                        
                    }
                    
                    digitalWrite(Ausgaenge_Array[tAusg_2], LOW);
                    digitalWrite(Ausgaenge_Array[tAusg_1], LOW);                
                    digitalWrite(Ausgaenge_Array[tAusg_1], HIGH);
                   
                    Eingang_An_Status[iii] = 2;

                    #ifdef DEBUG        // Serial output on 
                    Serial.println(" | Status: 2");
                    #endif
                }  
    
    
                
                //Mittelpunkt, beide Motoren ausschalten
                if(getPWM_Input > 1200 && getPWM_Input < 1800 && Eingang_An_Status[iii] != 0)
                {
                  
                    switch(iii)
                    {
                      case 0:
                        tAusg_1 = 0;
                        tAusg_2 = 1;
                       break;
                      case 1:
                        tAusg_1 = 2;
                        tAusg_2 = 3;
                       break;
                      case 2:
                        tAusg_1 = 4;
                        tAusg_2 = 5;
                       break;
                      case 3:
                        tAusg_1 = 6;
                        tAusg_2 = 7;
                       break;
                      case 4:
                        tAusg_1 = 8;
                        tAusg_2 = 9;
                       break;
                      default:
                        tAusg_1 = 0;
                        tAusg_2 = 1;
                       break;                                                        
                    }
                     
                    digitalWrite(Ausgaenge_Array[tAusg_1], LOW);
                    digitalWrite(Ausgaenge_Array[tAusg_2], LOW);
    
                    Eingang_An_Status[iii] = 0;
                    
                    #ifdef DEBUG        // Serial output on 
                    Serial.println(" | Status: Aus(0)");
                    #endif
                }
              }//Ende Abrage ob Eingang grösser 0 ist (aktiv)

          
        }//Ende For
        


      }//Ende While

     //Ende MultiUmpolmodul
  }
  else
  {       
       /*************************************
       * Beginn MultiServo
       *************************************/
       while(true)
       {
          //PWM Signal ab Empfänger einlesen
          getPWM_Input = pulseIn(PinIn2, HIGH);
        
        
          /*****************************************************************************
           * Aufruf der Taster Funktion für das Zählen des Ausgangs
           *****************************************************************************/
          if(getPWM_Input > 1600 && ErsterHighPuls == false)
            getCounter();
        
            //Kontrolle ob der Counter länger als x Sekunden nicht gedrückt wurde.
          /*****************************************************************************
           * Kontrollfunktion ob Ausgang gewählt ist (min. 1.5 kein neuer Impuls)
           *****************************************************************************/    
            if(counter > 0)
             CounterStatus();
        
        
          /*****************************************************************************
           * Der Taster kann so lange gedrückt bleiben, es wir nur immer 1 mal hochgezählt
           * Ist das PWM Signal in Ruhemodus, kann weiter hochgezählt werden oder die Freigabe für den Ausgang erfolgt
           * Hochzählen erfolgt nur auf positiven Ausschlage, negativer Ausschlag wird für das Rücksetzen benötigt.
           *****************************************************************************/
          if(getPWM_Input > 1800 && ErsterHighPuls == true)
              FreigabeAusgang = false;  //Freigabe nicht gegeben solange der Taster noch aktiv ist 
           else if(getPWM_Input > 1300 && getPWM_Input < 1800 && ErsterHighPuls == true)
          {
            alteZeit_PWM = 0;
            ErsterHighPuls = false;  
          }
        
        
          /*****************************************************************************
           * Counter und Ausgang Rücksetzen
           * Das Rücksetzen erfolgt über den negativ Ausschlag direkt und ohne Verzögerung
          *****************************************************************************/
        
          if(getPWM_Input < 1200)
            resetCounter();
        
        
          /*****************************************************************************
           * Aufruf Ausgang Schalt Funktion wenn der Taster bedient wurde und die gewünschte Zeit nicht mehr betätigt wurde.
           *****************************************************************************/
          if(counter > 0 && FreigabeAusgang == true)
            setAusgang();
       }//Ende While Schleife
  }//Ende MultiServo
  
}



/******************************************************************************
 * Kontroll Fukntion zur Kontorlle ob der Counter länger als 1.5 Sekunden nicht hochgezählt hat, 
 * erst dann wird Freigabe für PWM Signalweiterleitung gegeben
 *****************************************************************************/

 void CounterStatus()
 {
  aktZeit = millis();
  
  //Delta Zeit Berechnung
  if(alteZeit == 0)
     alteZeit= millis();
  else
    delta = aktZeit - alteZeit;

    if(delta > 1500)
      FreigabeAusgang = true;
    else
      FreigabeAusgang = false;

 }

/*****************************************************************************
Rücksetzen des Counters für den das erneute Ansteuer
*****************************************************************************/
void resetCounter()
{
      FreigabeAusgangReset = true; 
      FreigabeAusgang = false;
      SetInOutPWM = false;


      myservo.detach();        //Servo Ausgänge abhängen
      counter = 0;

}


/*****************************************************************************
 Einlesen des PWM Signals mit Erhöhung des Counters
*****************************************************************************/
void getCounter()
{

  if((millis() - alteZeit_PWM) > entPrell_PWM && alteZeit_PWM > 0)
  {
  
      //Counter erhöhen, ist die max. Ausgangsbelegung erreicht beginnt der Counter wieder von vorne
      if(counter == 10)
        counter = 1;
      else
        counter = counter + 1;

      //Erster Puls ist durch, index wurde hochgezählt
      ErsterHighPuls = true;

      alteZeit = alteZeit_PWM = millis();
        
  }
  else
    alteZeit_PWM = millis();    //Setzen der aktuellen zeit für die Berechnung der Freigabe des Ausgangs

}



/*****************************************************************************
Schalten des Ausganzg gemäss Counter Stand durch das hochzählen des Einganges
*****************************************************************************/
void setAusgang()
{

        //Counter beginng bei 1, erster Ausgang jedoch bei 3 => plus 2 damit der korrekte Ausgang aktiviert wird
        int setAusgang = counter+2;

       //Dauerschleife bis Counter Rücksetzung erfolgt
       while(counter > 0)
       {
          //Im ersten Durchlauf wird vor dem Einlesen der Ausgang als Servokanal deklariert
          if(SetInOutPWM == false)
          {
            //Aktiver Ausgang als Servo Ausgang setzen
            myservo.attach(setAusgang);
    
            getPWM_Output = pulseIn(PinIn13, HIGH);     //Einlesen erstes Signal
            myservo.writeMicroseconds(getPWM_Output);   //Schreiben des PWM Wertes
            SetInOutPWM = true;   
          }
          else if(SetInOutPWM == true)
          {
            getPWM_Output = pulseIn(PinIn13, HIGH);    //Einlesen erstes Signal
            myservo.writeMicroseconds(getPWM_Output);  //Schreiben des PWM Wertes
            
          }
        
            getPWM_Input = pulseIn(PinIn2, HIGH);      //Einlesen des Counter Signals, ändert sich dieses wird der Counter zurückgesetzt

            if(getPWM_Input < 1200)
              resetCounter();       

       }
       
}
